/*
    Copyright 2008 Luigi Auriemma

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA

    http://www.gnu.org/licenses/gpl.txt
*/

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>

#ifdef WIN32
#else
    #define stricmp strcasecmp
    #define stristr strcasestr
#endif

typedef uint8_t     u8;
typedef uint16_t    u16;
typedef uint32_t    u32;



#define VER         "0.1"
#define BUFFSZ      8192
#define WINZIP_KEY  "\x78\x56\x34\x12\x89\x67\x45\x23\x90\x78\x56\x34\x6E\xC9\x2F\x96"



int fgetz(u8 *data, int len, FILE *fd);
u8 hex2byte(u8 *hex);
int winzip_xflags(u8 *pwd);
void tea(u32 *p, u32 *key);



int main(int argc, char *argv[]) {
    FILE    *fd;
    u8      buff[BUFFSZ],
            *fname,
            *p;

    fputs("\n"
        "Winzip wjf xflags password decrypter "VER"\n"
        "by Luigi Auriemma\n"
        "e-mail: aluigi@autistici.org\n"
        "web:    aluigi.org\n"
        "\n", stdout);

    if(argc < 2) {
        printf("\n"
            "Usage: %s <file.WJF/xflags_password>\n"
            "\n", argv[0]);
        printf("\n  press RETURN to exit\n");
        fgetc(stdin);
        exit(1);
    }

    fname = argv[1];
    printf("- open %s\n", fname);
    fd = fopen(fname, "rb");
    if(!fd) {
        printf("- file not found, consider input as xflags password\n");
        sprintf(buff, "%.*s", sizeof(buff), fname);
        if(!winzip_xflags(buff)) printf("\n- password:\n  %s\n", buff);
        goto quit;
    }

    while(!fgetz(buff, sizeof(buff), fd)) {
    	p = strchr(buff, '=');
    	if(!p) continue;
    	*p++ = 0;

        if(!stricmp(buff, "xflags")) {
            if(!winzip_xflags(p)) printf("\n- password:\n  %s\n", p);
        }
    }

    fclose(fd);
quit:
    printf("\n  press RETURN to exit\n");
    fgetc(stdin);
    return(0);
}



int fgetz(u8 *data, int len, FILE *fd) {
    u8      *p;

    if(!fgets(data, len, fd)) return(-1);
    for(p = data; *p && (*p != '\n') && (*p != '\r'); p++);
    *p = 0;
    return(0);
}



u8 hex2byte(u8 *hex) {
    static const u8 hextable[256] =
        "\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        "\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        "\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        "\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\x00\x00\x00\x00\x00\x00"
        "\x00\x0a\x0b\x0c\x0d\x0e\x0f\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        "\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        "\x00\x0a\x0b\x0c\x0d\x0e\x0f\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        "\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        "\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        "\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        "\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        "\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        "\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        "\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        "\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        "\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00";

    return((hextable[hex[0]] << 4) | hextable[hex[1]]);
}



int winzip_xflags(u8 *pwd) {
    u32     len,
            pwdlen;
    u8      *p,
            *o,
            n1,n2,n3,n4;

    if(memcmp(pwd, "01", 2)) {
        printf("\nError: unsupported encrypted password method\n");
        pwd[0] = 0;
        return(-1);
    }

    for(p = pwd + 2, o = pwd; p[0] && p[1]; p += 8, o += 4) {
        n1 = hex2byte(p);
        n2 = hex2byte(p + 2);
        n3 = hex2byte(p + 4);
        n4 = hex2byte(p + 6);
        *(u32 *)o = (n1 << 24) | (n2 << 16) | (n3 << 8) | (n4);
    }
    len = o - pwd;
    if(len < 12) {
        printf("\nError: wrong xflags length (%d)\n", len);
        pwd[0] = 0;
        return(-1);
    }

    tea((void *)pwd, (void *)WINZIP_KEY);   // checksum
    pwdlen = *(u32 *)(pwd + 8);             // password length

    for(p = pwd + 12; p < o; p += 8) {      // password
        tea((void *)p, (void *)WINZIP_KEY);
    }

    if((pwdlen + 12) > len) pwdlen = len - 12;
    memmove(pwd, pwd + 12, pwdlen);
    pwd[pwdlen] = 0;
    return(0);
}



void tea(u32 *p, u32 *key) {
    u32     y,
            z,
            sum;
    int     i;

    y = p[0];
    z = p[1];
    sum = 0xc6ef3720;

    for(i = 0; i < 32; i++) {
        z -= (((y >> 5) ^ (y << 4)) + y) ^ (key[(sum >> 11) & 3] + sum);
        sum += 0x61c88647;
        y -= (((z >> 5) ^ (z << 4)) + z) ^ (key[sum & 3] + sum);
    }
    p[0] = y;
    p[1] = z;
}


